# -*- coding: utf-8 -*-

import os

from .mongo_proxy import get_mongo_proxy


class BaseConnection:
    """
    This takes care of connecting to db
    """

    DEFAULT_DB_NAME = 'db'

    @classmethod
    def init(cls, *args, **kwargs):
        """
        Setting up names for multiple db connections

        :param args:
            If provided, we expect single argument with name of the database

        :param kwargs:
            Possible to use only if args are unused
        """

        if args and len(args) == 1 and not kwargs:
            cls.configs = {cls.DEFAULT_DB_NAME: args[0]}
        elif args and kwargs or len(args) > 1:
            raise ValueError('Bad parameters. Use one arg or kwargs.')
        else:
            cls.configs = kwargs

        cls._masters = {}
        cls._masters_pids = {}
        cls._slaves = {}
        cls._slaves_pids = {}

    @classmethod
    def master(cls, db_name=None):
        """
        Returns connection to master/primary

        New connection for each worker
        If worker has connection already, return that

        :param db_name:
            Name of db we are connecting to
        """
        db_name = db_name or cls.DEFAULT_DB_NAME
        pid = os.getpid()
        if cls._masters.get(db_name) is None or cls._masters_pids.get(db_name) != pid:
            cls._masters[db_name] = cls.get_connection('MASTER', db_name=db_name)
            cls._masters_pids[db_name] = pid
        # log.debug('Connection MASTER {}, {}, {}, {}'.format(db_name, os.getpid(), pid, id(cls._masters[db_name])))
        return cls._masters[db_name]

    @classmethod
    def slave(cls, db_name=None):
        """
        Returns connection to slave/secondary

        New connection for each worker
        If worker has connection already, return that

        :param db_name:
            Name of db we are connecting to
        """
        db_name = db_name or cls.DEFAULT_DB_NAME
        pid = os.getpid()
        if cls._slaves.get(db_name) is None or cls._slaves_pids.get(db_name) != pid:
            cls._slaves[db_name] = cls.get_connection('SLAVE', db_name=db_name)
            cls._slaves_pids[db_name] = pid
        # log.debug('Connection SLAVE {}, {}, {} , {}'.format(db_name, os.getpid(), pid, id(cls._slaves[db_name])))
        return cls._slaves[db_name]

    @classmethod
    def get_connection(cls, prefix, db_name=None):
        """
        Abstract method which really creates the connection to db

        Classes inheriting from this class must implement this method

        :param prefix:
             MASTER / SLAVE

        :param db_name:
            Name of db we are connecting to
        """
        raise NotImplementedError()


class Connection(BaseConnection):
    """
    Create connection to MongoDB
    """

    read_prefs = {
        'MASTER': 'primary',
        'SLAVE': 'secondaryPreferred',
    }

    @classmethod
    def get_connection(cls, prefix, db_name=None):
        """
        Returns connction to master or slave - depending on prefix

        :param prefix:
            MASTER / SLAVE

        :param db_name:
            Name of db we are connecting to

        :return:
            Connection to db

        :rtype:
            :class:`mongo_proxy.MongoProxy`
        """
        return get_mongo_proxy(cls.configs[db_name or cls.DEFAULT_DB_NAME], client_kwargs={'readPreference': cls.read_prefs[prefix]})


class DbResources:
    """
    Crate for encapsulating resources neede for DB connection

    If you add new type of database, just add resources (cursor for MySql for example)
    """
    connection = None

    def __init__(self, connection=None):
        self.connection = connection



# -*- coding: utf-8 -*-

import re

from cerberus import Validator as BaseValidator


class Validator(BaseValidator):

    def _validate_type_objectid(self, value):
        if self.SCHEMA.get('nullable') and value is None:
            return True

        if re.match('[a-f0-9]{24}', str(value)):
            return True



# -*- coding: utf-8 -*-

from datetime import date, datetime, time
from json import JSONEncoder

from bson import ObjectId

# DATE_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'
DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'  # ISO 8601


def date_to_str(date_):
    """
    Converts a datetime value to the format defined in the configuration file.

    :param date: the datetime value to convert.
    """
    if isinstance(date_, str):
        return date_
    return datetime.strftime(date_, DATE_FORMAT) if date_ else None


def str_to_date(str_):
    """
    Converts a string to datetime

    :param str_: the string value to convert.
    """
    if isinstance(str_, datetime):
        return str_
    return datetime.strptime(str_, DATE_FORMAT) if str_ else None


class BaseJSONEncoder(JSONEncoder):

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, datetime):
            # convert any datetime to RFC 1123 format
            return date_to_str(obj)
        elif isinstance(obj, (time, date)):
            return obj.isoformat()
        elif isinstance(obj, set):
            return list(obj)
        return JSONEncoder.default(self, obj)


class MongoJSONEncoder(BaseJSONEncoder):

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, ObjectId):
            # BSON/Mongo ObjectId is rendered as a string
            return str(obj)
        else:
            # delegate rendering to base class method
            return super(MongoJSONEncoder, self).default(obj)



# -*- coding: utf-8 -*-

from slugify import slugify


def create_unique_slug(item, fields, db_conn, slug_key='slug', separator='-'):
    """
    Generates unique slug

    Checks if slug already exists in DB

    In `fields` must be keys to item or callables
    """

    i = 0
    tokens = []
    for field in fields:
        if callable(field):
            tokens.append(field(item))
        else:
            tokens.append(item[field])

    base_slug = slugify(separator.join(tokens), to_lower=True)
    slug = base_slug
    while db_conn.find_one({slug_key: slug}):
        i += 1
        slug = ''.join([base_slug, separator, str(i)])
    return slug



# -*- coding: utf-8 -*-

import hashlib
import logging as log
import random
import re
import string

from collections import Mapping
from copy import deepcopy
from sys import exc_info
from traceback import format_exception

from bson.json_util import dumps

from .encoders import MongoJSONEncoder

FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


def document_etag(value, ignore_fields=None):
    if ignore_fields:
        def filter_ignore_fields(doc, fields):
            for field in fields:
                key, _, value = field.partition(".")
                if value:
                    filter_ignore_fields(doc[key], [value])
                elif field in doc:
                    doc.pop(field)
                else:
                    pass

        value_ = deepcopy(value)
        filter_ignore_fields(value_, ignore_fields)
    else:
        value_ = value

    hashish = hashlib.sha1()
    json_encoder = MongoJSONEncoder()
    hashish.update(dumps(value_, sort_keys=True, default=json_encoder.default).encode('utf-8'))
    return hashish.hexdigest()


def camel_to_underscores(stringz0r):
    result = FIRST_CAP_RE.sub(r'\1_\2', stringz0r)
    return ALL_CAP_RE.sub(r'\1_\2', result).lower()


def random_string(length=8, alphabet=None):
    alphabet = alphabet or (string.ascii_letters + string.digits)
    return ''.join(random.choice(alphabet) for x in range(length))


def catch_exception_info():
    (type_exc, value, tbck) = exc_info()
    exc_str = ''.join(format_exception(type_exc, value, tbck))
    del tbck
    log.error(exc_str)
    return exc_str


def recursive_dict_update(dictionary, updates):
    """Recursively updates `dictionary` with `updates`"""
    for key, value in updates.items():
        if isinstance(value, Mapping):
            dictionary[key] = recursive_dict_update(dictionary.get(key, {}), value)
        else:
            dictionary[key] = value
    return dictionary



# -*- coding: utf-8 -*-


class QueryOptions:  # pylint: disable=too-few-public-methods
    """
    Crate with settings for db query
    """
    query = None
    sort = None
    page = 1
    limit = None
    projection = None
    id_field = None

    def __init__(self, query=None, sort=None, page=1, limit=None, projection=None, id_field=None):  # pylint: disable=too-many-arguments
        self.query = query
        self.sort = sort
        self.page = page
        self.limit = limit
        self.projection = projection
        self.id_field = id_field

    @property
    def offset(self):
        return (self.page - 1) * (self.limit or 0)

    def __repr__(self):
        return '<{} p:{} l:{} q:{} s:{}>'.format(self.__class__.__name__, self.page, self.limit, self.query, self.sort)



# -*- coding: utf-8 -*-

from copy import deepcopy
from datetime import datetime

from bson import ObjectId

from ..io.encoders import str_to_date
from ..io.slug import create_unique_slug
from ..io.utils import camel_to_underscores, document_etag
from ..io.validator import Validator


class BaseModel(Validator):

    ID_FIELD = '_id'
    CREATED_FIELD = '_created'
    UPDATED_FIELD = '_updated'
    ETAG_FIELD = '_etag'
    SLUG_FIELD = 'slug'
    SLUG_VALUE_FIELDS = None

    COLLECTION_NAME = None  # Defaults to class name without 'Model'

    SPECIAL_FIELDS = (CREATED_FIELD, UPDATED_FIELD, ETAG_FIELD)

    DEFAULT_FIELDS = {}

    SCHEMA = {
        ID_FIELD: {
            'type': 'objectid',
            'coerce': ObjectId,
        },
        CREATED_FIELD: {
            'type': 'datetime',
            'coerce': str_to_date,
        },
        UPDATED_FIELD: {
            'type': 'datetime',
            'coerce': str_to_date,
        },
        ETAG_FIELD: {
            'type': 'string',
        },
    }
    document = None

    def __init__(self, data=None, *args, **kwargs):
        schema = self.__get_schema_from_parents()
        schema.update(self.SCHEMA)
        self.SCHEMA = schema  # pylint: disable=invalid-name

        if 'schema' not in kwargs:
            kwargs['schema'] = schema

        super().__init__(*args, **kwargs)
        self.data = data or {}

    def __get_schema_from_parents(self):
        schema = {}
        for base in reversed(self.__class__.__bases__):
            if issubclass(base, BaseModel):
                schema.update(deepcopy(base.SCHEMA))
        return schema

    def __repr__(self):
        return '<{}: [{}] {}>'.format(self.__class__.__name__, self.get('_id'), self.get('slug') or self.get('name', ''))

    def __getitem__(self, key):
        source = self.data if self.document is None else self.document
        return source[key]

    def get(self, key, default=None):
        source = self.data if self.document is None else self.document
        return source.get(key, default)

    def validate(self, document=None, schema=None, update=False, normalize=True):
        return super().validate(document or self.data, schema=schema, update=update, normalize=normalize)

    def set_created_field(self):
        if self.CREATED_FIELD not in self.data:
            self.data[self.CREATED_FIELD] = datetime.utcnow()

    def set_updated_field(self):
        self.data[self.UPDATED_FIELD] = datetime.utcnow()

    def set_etag_field(self):
        self.data[self.ETAG_FIELD] = document_etag(self.data)

    def set_slug_field(self, ro_database):
        if self.SLUG_FIELD and self.SLUG_VALUE_FIELDS:
            self.data[self.SLUG_FIELD] = create_unique_slug(self.data, self.SLUG_VALUE_FIELDS, ro_database, slug_key=self.SLUG_FIELD)

    def set_default_fields(self):
        if self.DEFAULT_FIELDS:
            for key, value in self.DEFAULT_FIELDS.items():
                if key not in self.SCHEMA:
                    self.SCHEMA[key] = {'default': value}
                else:
                    self.SCHEMA[key]['default'] = value

    @classmethod
    def collection_name(cls):
        if cls.COLLECTION_NAME:
            return cls.COLLECTION_NAME
        else:
            return camel_to_underscores('{}{}'.format(cls.__name__.replace('Model', ''), cls.collection_suffix()))

    @classmethod
    def collection_suffix(cls):
        """
        Override to add suffix to collection name
        """
        return ''




from .db.connection import BaseConnection, Connection, DbResources
from .db.decorators import master, slave
from .db.mongo_proxy import get_mongo_proxy, MongoProxy
from .db.query_options import QueryOptions
from .io.encoders import DATE_FORMAT, MongoJSONEncoder, date_to_str, str_to_date
from .io.query_parser import QueryParser
from .io.slug import create_unique_slug
from .io.utils import camel_to_underscores, catch_exception_info, random_string
from .io.validator import Validator
from .managers.base import BaseManager, ValidationError
from .models.base import BaseModel



# -*- coding: utf-8 -*-

import time
import logging as log

import pymongo


def get_methods(*objs):
    return set(
        attr
        for obj in objs
        for attr in dir(obj)
        if not attr.startswith('_') and hasattr(getattr(obj, attr), '__call__')
    )

try:
    # will fail to import from older versions of pymongo
    from pymongo import MongoClient, MongoReplicaSetClient
except ImportError:
    MongoClient, MongoReplicaSetClient = None, None  # NOQA

EXECUTABLE_MONGO_METHODS = get_methods(
    pymongo.collection.Collection,
    pymongo.database.Database,
    pymongo.MongoReplicaSetClient,
    MongoClient,
    MongoReplicaSetClient,
    pymongo,
)


class Executable:  # pylint: disable=too-few-public-methods
    """
    Wrapper for Mongo methods, handling AutoReconnect exceptions
    """

    def __init__(self, method, logger, wait_time=None):
        self.method = method
        self.logger = logger
        self.wait_time = wait_time or 60

    def __call__(self, *args, **kwargs):
        """
        Handle AutoReconnect exception
        """
        start = time.time()
        i = 0
        while True:
            try:
                return self.method(*args, **kwargs)
            except pymongo.errors.AutoReconnect:
                end = time.time()
                delta = end - start
                if delta >= self.wait_time:
                    break
                self.logger.warning('AutoReconnecting, try {} ({} seconds)'.format(i, delta))
                time.sleep(pow(2, i))
                i += 1
        # Try one more time, but this time, if it fails, let the
        # exception bubble up to the caller.
        return self.method(*args, **kwargs)

    def __dir__(self):
        return dir(self.method)

    def __str__(self):
        return self.method.__str__()

    def __repr__(self):
        return self.method.__repr__()


class MongoProxy:  # pylint: disable=too-few-public-methods
    """
    Proxy for MongoDB connection.

    Called methods (find, insert, ..) are wrapped with :class:`Executable` so AutoReconnect exceptions are handled smoothly
    """

    def __init__(self, conn, logger=None, wait_time=None):
        """
        :param conn:
            Standard connection to db

        :param logger:
            Logger object

        :param wait_time:
            Timeout before reconnecting after AutoReconnect exception
        """
        if logger is None:
            # import logging
            # logger = logging.getLogger(__name__)
            logger = log.getLogger(__name__)

        self.conn = conn
        self.logger = logger
        self.wait_time = wait_time

    def __getitem__(self, key):
        """
        Create and get proxy for method, called with connection named `key`

        :param key:
            Connection name
        """
        item = self.conn[key]
        if hasattr(item, '__call__'):
            return MongoProxy(item, self.logger, self.wait_time)
        return item

    def __getattr__(self, key):
        """
        If `key` is name of executable method (find, insert, ..),
        this method is wrapped to instance of :class:`Executable`
        """

        attr = getattr(self.conn, key)
        if hasattr(attr, '__call__'):
            if key in EXECUTABLE_MONGO_METHODS:
                return Executable(attr, self.logger, self.wait_time)
            else:
                return MongoProxy(attr, self.logger, self.wait_time)
        return attr

    def __call__(self, *args, **kwargs):
        return self.conn(*args, **kwargs)

    def __dir__(self):
        return dir(self.conn)

    def __str__(self):
        return self.conn.__str__()

    def __repr__(self):
        return self.conn.__repr__()

    def __nonzero__(self):
        return True


def get_mongo_proxy(config, client_kwargs=None):
    """
    Returns Mongo Proxy

    :param config:
        Configi with connection details

        Must include:

        MONGO_USERNAME
        MONGO_PASSWORD
        MONGO_HOST
        MONGO_PORT
        MONGO_DBNAME

    :type config:
        dict

    :param client_kwargs:
        Additional config passed to :class:`pymongo.MongoClient` constructor during connection creation

    :type client_kwargs:
        dict
    """
    client_kwargs = client_kwargs or {}

    conn_str = 'mongodb://{}:{}@{}:{}/{}'.format(
        config['MONGO_USERNAME'],
        config['MONGO_PASSWORD'],
        config['MONGO_HOST'],
        config['MONGO_PORT'],
        config['MONGO_DBNAME'],
    )

    if config.get('MONGO_SOCKET_TIMEOUT'):
        client_kwargs['socketTimeoutMS'] = int(config['MONGO_SOCKET_TIMEOUT'])

    if config.get('MONGO_CONNECT_TIMEOUT'):
        client_kwargs['connectTimeoutMS'] = int(config['MONGO_CONNECT_TIMEOUT'])

    if config.get('MONGO_SERVER_SELECTION_TIMEOUT'):
        client_kwargs['serverSelectionTimeoutMS'] = int(config['MONGO_SERVER_SELECTION_TIMEOUT'])

    log.info('Mongo Connection: %s@%s:%s/%s %s', config['MONGO_USERNAME'], config['MONGO_HOST'], config['MONGO_PORT'], config['MONGO_DBNAME'], client_kwargs)
    client = MongoClient(conn_str, **client_kwargs)
    return MongoProxy(client[config['MONGO_DBNAME']])



# -*- coding: utf-8 -*-

from collections import defaultdict
from copy import deepcopy
import logging as log

import pymongo

from ..db.decorators import master, slave
from ..db.query_options import QueryOptions
from ..models.base import BaseModel


class ValidationError(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errors = None


def query_log(self, *args):
    log.debug(
        '%s(%s): %s',
        self.__class__.__name__,
        self.MODEL_CLASS.__name__,
        ', '.join([str(arg) for arg in args]))


class BaseManager:

    MODEL_CLASS = BaseModel
    DEFAULT_SORT = MODEL_CLASS.ID_FIELD
    # Query which is applied always
    BASE_QUERY = {}

    connection = None

    @classmethod
    def factory(cls, model_class, default_sort=None, base_query=None):
        """
        Factory method so we don't need to subclass every time
        """
        instance = cls()
        instance.MODEL_CLASS = model_class  # pylint: disable=invalid-name
        instance.DEFAULT_SORT = default_sort or model_class.ID_FIELD  # pylint: disable=invalid-name
        instance.BASE_QUERY = base_query or {}  # pylint: disable=invalid-name
        return instance

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    @master
    def rw_conn(self):
        return self.connection[self.MODEL_CLASS.collection_name()]

    @slave
    def ro_conn(self):
        return self.connection[self.MODEL_CLASS.collection_name()]

    def get(self, options):
        if isinstance(options, QueryOptions):
            pass
        elif isinstance(options, dict):
            options = QueryOptions(
                query=options
            )
        else:
            item_id = options
            options = QueryOptions(
                query={self.MODEL_CLASS.ID_FIELD: item_id}
            )

        self.__apply_base_query(options.query)

        model = self.MODEL_CLASS(options.query)
        model = self.__remove_defaults(model, options.query)

        # if not model.validate(update=True):
        #    self._validation_error(model)
        # Ignore validation errors for query
        model.validate(update=True)
        sort = self.__transform_sort(options.sort or self.DEFAULT_SORT)

        query_log(self, model.document)
        item = self.ro_conn().find_one(model.document, projection=options.projection, sort=sort)

        if item:
            instance = self.MODEL_CLASS(item)
            if not instance.validate(update=bool(options.projection)):
                self._log_validation_message(instance)
            return instance.data

    @staticmethod
    def __remove_defaults(model, query):
        """
        Remove default values generated for some fields.
        We don't want them to get into the query and modify it.
        It brings evil.
        """

        query = query or {}
        keys_to_remove = []
        for key, schema in model.schema.items():
            if key not in query and 'default' in schema:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            del model.schema[key]
        return model

    def find(self, options=None):

        options = options or QueryOptions()

        if not isinstance(options, QueryOptions):
            raise TypeError('Options must be QueryOptions instance')

        model = self.MODEL_CLASS(options.query)
        model = self.__remove_defaults(model, options.query)

        # Ignore validation errors for query
        model.validate(update=True)

        limit = options.limit
        offset = options.offset or 0
        projection = options.projection
        sort = self.__transform_sort(options.sort or self.DEFAULT_SORT)

        self.__apply_base_query(model.document)
        query_log(self, model.document)
        query = self.ro_conn().find(model.document, projection=projection)

        if sort:
            query = query.sort(sort)

        if limit is not None:
            query = query.skip(offset).limit(limit)

        valid_items = []

        for item in query:
            instance = self.MODEL_CLASS(item)
            instance.validate()
            valid_items.append(instance.data)

            if instance.errors:
                # If projection is present, required fields are likely to be missing
                if projection:
                    for key in projection:
                        if key in instance.errors:
                            self._log_validation_message(instance)
                            break
                else:
                    self._log_validation_message(instance)

        return valid_items

    @classmethod
    def __apply_base_query(cls, query):
        """
        Adds missing attributes to query from cls.BASE_QUERY
        """
        for key, value in cls.BASE_QUERY.items():
            if key not in query:
                query[key] = value

    def _log_validation_message(self, instance):
        log.debug('Model Validation: %s [%s]', self.__class__.__name__, instance.get(self.MODEL_CLASS.ID_FIELD))

    @staticmethod
    def __transform_sort(sort):
        transformed = []
        items = [item.strip() for item in sort.split(',')]
        for item in items:
            order = pymongo.DESCENDING if item.startswith('-') else pymongo.ASCENDING
            field = item.replace('-', '')
            transformed.append((field, order))
        return transformed

    @classmethod
    def _validation_error(cls, model):
        msg = '{} is not valid'.format(cls.MODEL_CLASS.__name__)
        log.warning('%s. ERRORS: %s', msg, model.errors)
        log.warning('ERROR VALUES: %s', {key: model.document.get(key) for key in model.errors})
        err = ValidationError(msg)
        err.errors = model.errors
        raise err

    def create(self, data):
        if isinstance(data, self.MODEL_CLASS):
            model = data
        else:
            model = self.MODEL_CLASS(data)

        model.set_created_field()
        model.set_updated_field()
        model.set_etag_field()
        model.set_default_fields()

        if self.MODEL_CLASS.SLUG_FIELD not in model.data:
            model.set_slug_field(self.ro_conn())

        if not model.validate():
            self._validation_error(model)

        query_log(self, model.document)
        result = self.rw_conn().insert_one(model.document)
        model.document.update({self.MODEL_CLASS.ID_FIELD: result.inserted_id})
        model.data.update({self.MODEL_CLASS.ID_FIELD: result.inserted_id})
        return model.data

    def update(self, options, data, many=False, upsert=False, operation='$set', modifiers=None):  # pylint: disable=too-many-arguments
        modifiers = modifiers or {}
        model = self.MODEL_CLASS(data)
        model = self.__remove_defaults(model, data)

        model.set_updated_field()
        model.set_etag_field()

        if upsert:
            model.set_created_field()
            if self.MODEL_CLASS.SLUG_FIELD not in model.data:
                model.set_slug_field(self.ro_conn())

        if not model.validate(update=not upsert):
            self._validation_error(model)

        query, is_many = self._get_query_and_strategy(options)

        if is_many or many:
            update_fn = self.rw_conn().update_many
        else:
            update_fn = self.rw_conn().update_one

        update_query = self._incorporate_special_fields(model, operation, modifiers)

        query_log(self, query, update_query)
        result = update_fn(query, update_query, upsert=upsert)
        return result.modified_count

    @classmethod
    def _incorporate_special_fields(cls, model, operation, modifiers):
        """
        In case we want to make $push, it migh occur that fields like _update are added too. That means we are fucked.
        We must take care of that so we do $push and $set in one query.
        """
        query = defaultdict(dict)
        if operation == '$set':
            query['$set'] = model.document
        else:
            for field in model.SPECIAL_FIELDS:
                if field in model.document:
                    query['$set'][field] = cls._apply_modifier_dude(deepcopy(model.document), deepcopy(modifiers), key=field)
                    model.document.pop(field)
                    query[operation] = cls._apply_modifier_dude(deepcopy(model.document), deepcopy(modifiers))
        return query

    @classmethod
    def _apply_modifier_dude(cls, document, modifiers, key=None):
        if key:
            if key in modifiers:
                modifier = modifiers.pop(key)
                return modifier(document[key])
                # return {modifiers.pop(key): document[key]}
            else:
                return document[key]

        doc = {}
        for key, value in document.items():
            if key in modifiers:
                modifier = modifiers.pop(key)
                doc[key] = modifier(value)
            else:
                doc[key] = value
        return doc

    def _get_query_and_strategy(self, options):
        """
        Decides how query will look like and if we are working with single or many docs.
        """
        is_many = False
        if isinstance(options, QueryOptions):
            query = options.query
            is_many = True
        elif isinstance(options, dict):
            model = self.MODEL_CLASS(options)
            model = self.__remove_defaults(model, options)
            model.validate(update=True)
            query = model.document
        else:
            model = self.MODEL_CLASS({self.MODEL_CLASS.ID_FIELD: options})
            model = self.__remove_defaults(model, {self.MODEL_CLASS.ID_FIELD: options})
            model.validate(update=True)
            query = model.document
        return query, is_many

    def delete(self, options):
        query, is_many = self._get_query_and_strategy(options)

        if is_many:
            delete_fn = self.rw_conn().delete_many
        else:
            delete_fn = self.rw_conn().delete_one

        query_log(self, query)
        result = delete_fn(query)
        return result.deleted_count

    def count(self, options=None):
        options = options or QueryOptions()

        if not isinstance(options, QueryOptions):
            raise TypeError('Options must be QueryOptions instance')

        self.__apply_base_query(options.query)
        model = self.MODEL_CLASS(options.query)
        model = self.__remove_defaults(model, options.query)

        # Ignore validation errors for query
        model.validate(update=True)

        query_log(self, model.document)
        return self.ro_conn().count(model.document)




from .encoders import DATE_FORMAT
from .utils import recursive_dict_update

FILTER_OPS = {
    'eq': '$eq',
    'is': '$eq',
    'ne': '$ne',
    'isnt': '$ne',
    'gt': '$gt',
    'gte': '$gte',
    'lt': '$lt',
    'lte': '$lte',
    'in': '$in',
    'nin': '$nin',
    're': '$regex',
}

class QueryParser:

    @staticmethod
    def _get_query(key, opval, coerce_fn):
        """Parses token and returns subquery based on operator inside token

        :param key      - name of the field
        :param opval   - string in format operator.value, operator is optional, defaults to eq

        Supported operators vs Mongo equivalent:
            eq  - $eq, default
                filter=animal:eq.cat
                filter=animal:cat - alternative notation
            ne  - $ne - complementary operator for eq
            gte - $gte
            lte - $lte
                filter=_created:gte.2018-02-12T14:26:09.666000Z
            gt  - $gt
            lt  - $lt
                filter=_created:lt.2018-02-12T14:26:09.666000Z
            is  - $eq - in this case 'null' is converted to None, 'false' to False, 'true' to True
                filter=backup:is.null
            isnt- $ne - complementary operator for is
            in  - $in - use '|' as separator
                filter=status:in.running|canceled
            nin  - $nin - complementary operator for in
                filter=status:nin.running|pending
            re  - $regex - searches by regular expression
                filter=status:re.*something*

        """
        if not isinstance(opval, str):
            return opval

        tokens = opval.split('.', 1)
        # Ve have value only - use default eq operator
        if len(tokens) == 1:
            return {key: coerce_fn(opval)}

        operator, val = tokens

        if operator in ('in', 'nin',):
            val = [coerce_fn(value) for value in val.split('|')]
        elif operator in ('is', 'isnt'):
            lval = val.lower()
            # Value transformation
            if lval in ('true', 'false'):
                val = lval == 'true'  # Hipster transformation to bool
            elif lval in ('null', 'none'):
                val = None
            else:
                raise ValueError('Query: {}.{} is invalid - invalid value for such operation'.format(operator, val))
        else:
            val = coerce_fn(val)

        return {key: {FILTER_OPS[operator]: val}}

    @classmethod
    def parse_filter(cls, query_str, coerce_fn):
        """Parses ?filter=. Raises ValueError"""
        query = {}
        if query_str:
            tokens = query_str.split(';')  # Use ; as separator

            for token in tokens:
                key, opval = token.split(':', 1)
                subquery = cls._get_query(key, opval, coerce_fn(key))
                if not subquery:
                    raise ValueError('Invalid query: {}'.format(token))
                recursive_dict_update(query, subquery)
        return query

    @staticmethod
    def parse_limit(query_str):
        """Parses ?limit=. Raises ValueError"""
        query_str = query_str or 0
        return int(query_str) or None

    @staticmethod
    def parse_sort(query_str):
        """Parses ?sort=. Raises ValueError"""
        if query_str:
            return query_str.replace(';', ',')  # Allow both (;,) as separator

    @staticmethod
    def parse_projection(query_str):
        """Parses ?fields=. Raises ValueError"""
        if query_str:
            return query_str.replace(';', ',').split(',')  # Allow both (;,) as separator



# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages


def __path(filename):
    return os.path.join(os.path.dirname(__file__), filename)

with open(__path('requirements.txt')) as f:
    REQUIRES = [l.strip() for l in f.readlines()]

VERSION = '0.2.4'

setup(
    name='ruin',
    packages=find_packages(),
    install_requires=REQUIRES,
    version=VERSION,
    description='RUIN - Python/Mongo query builder and model validator',
    author='g----',
    author_email='gloob@krutt.org',
    url='https://gitlab.com/g----/ruin',
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
)



# -*- coding: utf-8 -*-

import logging as log
from functools import wraps
from time import sleep

# from pymongo.errors import ServerSelectionTimeoutError, NotMasterError  # pymongo 3+

from .connection import Connection, DbResources


def transaction(db_name=None, master_=True):
    """
    This function returns decorator that provides access to db connection in decorated method (or function)

    :param db_name:
        Name of db we are connecting to

    :param master_:
        Flag to create either master or slave
    :type master_:
        bool
    """

    def decorator(function):

        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            # except (ServerSelectionTimeoutError, NotMasterError) as exc:  # pymongo 3+
            except Exception as exc:  # pylint: disable=broad-except
                log.fatal('MONGO FATAL ERROR: %s', repr(exc))
                sleep_seconds = 0.1
                log.info('SLEEPING FOR %s SECONDS BEFORE RETRY', sleep_seconds)
                sleep(sleep_seconds)
                return function(*args, **kwargs)

        @wraps(function)
        def method_wrapper(self, *args, **kwargs):
            """
            This is used when decorating method inside class
            """
            _db_name = db_name or getattr(self, 'DATABASE_NAME', None)
            self.connection = Connection.master(_db_name) if master_ else Connection.slave(_db_name)
            return wrapper(self, *args, **kwargs)

        @wraps(function)
        def function_wrapper(*args, **kwargs):
            """
            This is used when decorating function
            """
            connection = Connection.master(db_name) if master_ else Connection.slave(db_name)
            return wrapper(DbResources(connection=connection), *args, **kwargs)
        return method_wrapper if callable(function) else function_wrapper
    return decorator

master = transaction(master_=True)  # pylint: disable=invalid-name
slave = transaction(master_=False)  # pylint: disable=invalid-name
